************************************************************************
*                                                                      *
*  FUNCTION UnPackBytes(    BufPtr : POINTER;                          *
*                           BufSize: INTEGER;                          *
*                           VAR PicPtr : POINTER;                      *
*                           VAR PicSize: INTEGER) : INTEGER;           *
*                                                                      *
*  Params:  BufPtr - pointer to area containing packed data            *
*           BufSize- max size of area containing packed data           *
*           PicPtr - Pointer to start of area to unpack into           *
*           PicSize- Maximum space available for unpacked picture      *
*                                                                      *
*  Returns: UnPackBytes return Value: Actual number of bytes unpacked  *
*           PicPtr:    Updated to one past end of last byte unpacked   *
*           PicSize: reduced by number of bytes of picture generated   *
*                                                                      *
*        ------------------------------------------------------        *
*                                                                      *
*  Sample Calls:                                                       *
*                                                                      *
*        OpenFile;                                                     *
*        Mark:=StartofFile;                                            *
*        PicSize:=$7d00;                                               *
*        BufSize:=$
;                                                *
*        REPEAT                                                        *
*          SetFileMark(Mark);                                          *
*          Read(f,BufPtr,Bufsize);                                     *
*          HowMuch:=UnPackBytes(BufPtr,BufSize,PicPtr,PicSize);        *
*          Mark:=Mark+HowMuch;                                         *
*        UNTIL ( (PicSize=0) OR EOF(f) );                              *
*                                                                      *
*     -----------------------------------------------------------      *
*                                                                      *
*    Packed data incription format:                                    *
*                                                                      *
*        %00xxxxxx : (xxxxxx : 0->63) =    1 to 64 bytes follow -      *
*               All Different                                          *
*                                                                      *
*        %01xxxxxx : (xxxxxx : 1,2,3,5,6,7) = 1,2,3,5,6,7 repeats -    *
*               of next byte                                           *
*                                                                      *
*        %10xxxxxx : (xxxxxx : 0->63) =    1 to 64 repeats -           *
*               of next 4 bytes                                        *
*                                                                      *
*        %11xxxxxx : (xxxxxx : 0->63) =    1 to 64 times 4 repeats -   *
*               of next byte                                           *
*                                                                      *
*    -----------------------------------------------------------       *

           use   4/Locator.Macs
           use   4/Mem.Macs
           use   4/Misc.Macs
           use   4/Util.Macs

           ORG   $038000
           MX    %00

**************************

           PushWord  #$0000    ; Space For Result
           PushLong  Buf       ; Buf     04/0000 Packed Data
           PushWord  #$5B8D    ; BufSize         Packed Data Size
           PushLong  #Pic      ; Pic Ptr 05/0000
           PushLong  #PicSize  ; Pic size Ptr

           JSL   Begin
           RTL
           
Begin      JSL   UNPACKBYTES
           RTL

Buf        HEX   00,00,04,00    ; $04/0000
Pic        HEX   00,00,05,00    ; $05/0000
PicSize    HEX   00,80          ; $8000

**************************************
* Main UnPackbits entry              *
**************************************

FROM        EQU  1              ; LONG
TO          EQU  5              ; LONG   FROM+4
ENDBUF      EQU  9              ; LONG   TO+4
TEMP        EQU  13             ; LONG   ENDBUF+4
COUNT       EQU  17             ; WORD   TEMP+4
LF          EQU  19             ; WORD   COUNT+2

ORIGDRCT    EQU  21             ; WORD   LF+2

RTL1        EQU  23             ; ADDR3  ORIGDRCT+2
RTL2        EQU  26             ; ADDR3  RTL1+3

PICSIZEP    EQU  29             ; LONG   RTL2+3
PICPTRP     EQU  33             ; LONG   PICSIZEP+4   PARAM : BufPtr     4 bytes
BUFSIZE     EQU  37             ; WORD   PICPTRP+4    PARAM : BufSize    2 bytes
BUFPTR      EQU  39             ; LONG   BUFSIZE+2    PARAM : PicPtr     4 bytes
RESULT      EQU  43             ; WORD   BUFPTR+4     PARAM : PicSize    2 bytes
                                ;      -> +2

UNPACKBYTES PHD                 ; SET UP STACK AS ZERO PAGE +PARMS
            TSC
            SEC
            SBC  #ORIGDRCT-1    ; 20
            TCS
            TCD

            LDA  [PICPTRP]      ; GET THE ARGS.
            TAX                 ; (PICPTRP NEEDS DE-REFERANCE)
            LDY  #2
            LDA  [PICPTRP],Y
            STA  TO+2
            STX  TO

            LDA [PICSIZEP]
            STA COUNT

            LDA  BUFPTR
            STA  FROM
            CLC
            ADC  BUFSIZE
            STA  ENDBUF
            LDA  BUFPTR+2
            STA  FROM+2
            ADC  #0
            STA  ENDBUF+2

DOUNPACK    LDY  #0
            LDA  [FROM],Y      ; GET THE COMMAND
            STA  TEMP
            AND  #$3F
            TAX                ; GET COUNT ONLY INTO X
            INC  FROM          ; BUMP TO FIRST BYTE TO TRANSFER
            BNE  M1
            INC  FROM+2
M1          BIT  TEMP-1        ; V & C BITS GIVE COMMAND
            BMI  REP4
            BVS  REP1          ; 01xxxxxx -> REPEAT NEXT 1 BYTE
            BRA  DIFF          ; 00xxxxxx -> ALL DIFFERENT
REP4        BVS  REP41         ; 11xxxxxx -> REPEAT 4 OF NEXT 1 BYTE
            BRA  REP44         ; 10xxxxxx -> REPEAT OF NEXT 4 BYTES

OUTP        PLA
            JMP  UDONE         ; GO FINISH UP

*****************************************
*                                       *
*   DIFF : 0 ->63 IN A MEANS  1->64     *
*   DIFFERENT BYTES FOLLOW TO BE MOVED  *
*                                       *
*****************************************
DIFF        PHA                ; WHEN WE GO TO COMMON CODE, WE NEED
                               ; SOMETHING STACKED (OTHER CALLS USE IT)
            BRA  JOIN

*****************************************
*                                       *
* REP1 : 1,2,3,5,6,7 REPEATS OF FOLLOW- *
*        ING BYTE.                      *
*                                       *
*****************************************
REP1        LDA  TEMP           ; PUT BYTE IF LOW & HIGH & ON STACK
            AND  #$FF00
            PHA
            XBA
            ORA  1,S
            STA  1,S
            BRA  JOIN           ; JOIN UP, X=LEN IS CORRECT STILL

*****************************************
*                                       *
* REP41 : IS SAME AS REP1 ABOVE, EXCEPT *
*         COUNT IN X MUST BE 4 TIMES >  *
*                                       *
*****************************************
REP41       INX
            TXA
            ASL A
            ASL A
            TAX
            DEX
            BRA REP1

*****************************************
*                                       *
* REP44 : MUST SET UP FOR 2 STORES I.E. *
*         4 BYTES WORTH, AND COUNT IS   *
*         DOUBLED (2 STORES EACH TIME)  *
*                                       *
*****************************************
REP44       INX
            TXA
            ASL  A
            TAX                 ; FIRST DOUBLE X
            DEX
            PHA                 ; MUST PUT SOMETHING ON STACK FOR
                                ; OTHER CASES

* BRA   JOIN                    ; WE FIND 4 BYTES AT [FROM],
* (FALL THROUGH TO JOIN)        ; SO MUST REMEMBER TO FIND IT THERE

*****************************************
*                                       *
* JOIN : HERE, THE LOAD STORE SEQUENCE  *
*        OF DATA TRANSFER TAKES PLACE   *
*                                       *
*****************************************
JOIN        CPX  COUNT          ; BE SURE THERE'S ROOM
            BGE  OUTP           ; NO - DONE   (BUT PULL OFF EXTRA PUSH)

            LDA  #1             ; CHECK IF LENGTH WOULD GO OVER
            BIT  TEMP-1         ; DEPENDS ON TYPE OF CMD.
            BVS  L1             ; LEN=1 FOR %01 & %11
            BMI  L4             ; LEN=4 FOR %10
            TXA
            INC  A              ; LEN=N+1 FOR %00..N...
            BRA  L1
L4          LDA  #4
L1          CLC                 ; FROM+LEN .LT. ENDBUF MEANS NOT DONE
            ADC  FROM
            PHA                 ; SAVE FOR COMPARE
            LDA  FROM+2
            ADC  #0
            CMP  ENDBUF+2
            BCC  NOTDONP
            PLA
            CMP  ENDBUF
            BLT  NOTDON
            BEQ  NOTDON
            BRA  OUTP
NOTDONP     PLA

NOTDON      CPX  #0             ; (NEEDED FOR 1ST ENTRY IN LOOP)
LOOP        BMI  DOUPD          ; YES

*****************************************
*                                       *
* MOVEIT : USED LOCALLY TO FETCH PROPER *
*          INPUT, DEPENDING ON TYPE OF  *
*          CMD. BEING PROCESSED         *
*  NOTE: X<0 MEANS STORE 1 BYTE ONLY    *
*****************************************
            BIT  TEMP-1         ; THIS HAS THE FLAGS
            BMI  MREP4
            BVS  MREP1
            CPX  #0             ; IF NEG. IT'S ONE BYTE ONLY
            BEQ  MSPEC
            LDA  [FROM],Y
            STA  [TO],Y
            INY
            INY
            JMP  NDMOV
MSPEC       SEP  #%00100000
            LDA  [FROM],Y
            STA  [TO],Y
            REP  #%00100000
            INY
            JMP  NDMOV
MREP1       CPX  #0
            BEQ  M2SPEC
            LDA  1,S
            STA  [TO],Y
            INY
            INY
            JMP  NDMOV
M2SPEC      SEP  #%00100000
            LDA  1,S
            STA  [TO],Y
            REP  #%00100000
            INY
            JMP  NDMOV
MREP4       BVS  MREP1          ; REP41 SAME AS REP1    BUT LARGER COUNT
            LDA  [FROM]
            STA  [TO],Y
            PHY                 ; HOLD Y A SEC
            LDY  #2
            LDA  [FROM],Y
            PLY
            INY
            INY
            STA  [TO],Y
            INY
            INY
******
NDMOV       DEX
            DEX
            BRA  LOOP           ; GO DO NEXT ONES
DOUPD       PLA                 ; GET RID OF EXTRA PUSH
            LDA  TEMP           ; UPDATE APPROPRIATE POINTERS
            AND  #$3F
            INC  A
            BIT  TEMP-1
            BMI  XREP4          ; FROM DEPENDS ON TYPE
            BVC  XDIFF          ; FOR 00xxxxxx ADD LENGTH
FP1         LDA  #1             ; FOR 01xxxxxx ADD 1
XDIFF       CLC
            ADC  FROM
            STA  FROM
            LDA  FROM+2
            ADC  #0
            STA  FROM+2
            BRA  DOTO
XREP4       BVS  FP1            ; FOR 11xxxxxx ALSO    ADD 1
            LDA  #4             ; FOR 10xxxxxx ADD 4
            BRA  XDIFF

DOTO        TYA                 ; Y IS OUTPUT COUNT
            CLC
            ADC  TO
            STA  TO
            LDA  TO+2
            ADC  #0
            STA  TO+2
            TYA
            STA  TEMP
            LDA  COUNT
            SEC
            SBC  TEMP
            STA  COUNT
            JMP  DOUNPACK

UDONE       LDA  FROM
            SEC
            SBC  BUFPTR
            DEC  A
            STA  RESULT         ; SET RESULT LENGTH

            LDA  TO
            SEC
            SBC  [PICPTRP]
            PHA                 ; HOLD IT, NEED SIZE - IT
            LDA  [PICSIZEP]
            SEC
            SBC  1,S
            STA  [PICSIZEP]
            PLA

            LDA  TO
            STA  [PICPTRP]
            LDA  TO+2
            LDY  #2
            STA  [PICPTRP],Y

            LDA  ORIGDRCT
            TCD
            LDA  RTL1,S         ; FIX STACK UP CORRECTLY
            STA  RESULT-6,S
            LDA  RTL1+2,S
            STA  RESULT-4,S
            LDA  RTL1+4,S
            STA  RESULT-2,S
            TSC
            CLC
            ADC  #RESULT-7
            TCS
            LDA  #0             ; fix by SEG 5 March 87
            CLC                 ; fix by SEG 5 March 87
            RTL

***********************************************************************


